import re
import itertools
from datetime import datetime, date
import pytz
from pandas.tseries.offsets import BDay


def listify(value):
    """
    Convert the input to a list if it's not already a list.

    Args:
    value: Any type of input

    Returns:
    list: The input as a list
    """
    if isinstance(value, list):
        return value
    return [value]


def is_valid_date(date_str: str, date_format: str = "%Y-%m-%d") -> bool:
    """
    Check if the date is valid regarding the specified date format
    """
    pattern = r"^\d{4}-\d{2}-\d{2}$"
    if not re.match(pattern, date_str):
        return False
    try:
        datetime.strptime(date_str, date_format)
        return True
    except ValueError:
        return False


def is_valid_filing_date(filing_date: str) -> bool:
    """
    Check if filing_date is given in the correct format
    """
    if ":" in filing_date:
        # Check for only one colon
        if filing_date.count(":") > 1:
            return False
        start_date, end_date = filing_date.split(":")
        if start_date:
            if not is_valid_date(start_date):
                return False
        if end_date:
            if not is_valid_date(end_date):
                return False
    else:
        if not is_valid_date(filing_date):
            return False
    return True


def is_start_of_quarter():
    """
    Check if today is the first business day of the quarter
    """
    today = datetime.now().date()

    # Check if it's the start of a quarter
    if today.month in [1, 4, 7, 10] and today.day <= 5:
        # Get the first day of the current quarter
        first_day_of_quarter = datetime(today.year, today.month, 1).date()

        # Calculate one business day after the start of the quarter
        one_business_day_after = (first_day_of_quarter + BDay(1)).date()

        # Check if we haven't passed one full business day yet
        if today <= one_business_day_after:
            return True
    return False


def split_filing_date(filing_date: str) -> tuple[date, date]:
    """
    Get start and end date from filing date
    """
    if ":" in filing_date:
        start_date, end_date = filing_date.split(":")

        if not start_date:
            start_date = date(1994, 6, 1)
        else:
            start_date = datetime.strptime(start_date, "%Y-%m-%d").date()

        if not end_date:
            end_date = date.today()
        else:
            end_date = datetime.strptime(end_date, "%Y-%m-%d").date()

    else:
        start_date = datetime.strptime(filing_date, "%Y-%m-%d").date()
        end_date = start_date

    return start_date, end_date


def filing_date_to_year_quarters(filing_date: str) -> list[tuple[int, int]]:
    """
    Build list of year and quarter combinations from timeframe in filing_date
    """
    start_date, end_date = split_filing_date(filing_date)
    start_quarter = (start_date.month - 1) // 3 + 1
    end_quarter = (end_date.month - 1) // 3 + 1

    result = []
    for year in range(start_date.year, end_date.year + 1):
        if year == start_date.year and year == end_date.year:
            quarters = range(start_quarter, end_quarter + 1)
        elif year == start_date.year:
            quarters = range(start_quarter, 5)
        elif year == end_date.year:
            quarters = range(1, end_quarter + 1)
        else:
            quarters = range(1, 5)

        for quarter in quarters:
            result.append((year, quarter))
    return result


def current_year_and_quarter() -> tuple[int, int]:
    """
    Get the year and quarter of today's date
    """
    # Define the Eastern timezone
    eastern = pytz.timezone("America/New_York")

    # Get the current time in Eastern timezone
    now_eastern = datetime.now(eastern)

    # Calculate the current year and quarter
    current_year, current_quarter = now_eastern.year, (now_eastern.month - 1) // 3 + 1
    return current_year, current_quarter


def available_quarters() -> list[tuple[int, int]]:
    """
    Get a list of year and quarter tuples
    :return:
    """
    current_year, current_quarter = current_year_and_quarter()
    start_quarters = [(1994, 3), (1994, 4)]
    in_between_quarters = list(itertools.product(range(1995, current_year), range(1, 5)))
    end_quarters = list(itertools.product([current_year], range(1, current_quarter + 1)))
    return start_quarters + in_between_quarters + end_quarters



import os.path
from pathlib import Path

from pydantic import BaseModel
from bs4 import BeautifulSoup
import pandas as pd

from nport._connector import download_text
from nport.fund import Registrant, Fund
from nport.instruments import BaseInstrument, DebtSecurity, Derivative


BASE_DIR, _ = os.path.split(__file__)


class NPORT(BaseModel):
    """
    Class of a NPORT-P Filing
    """
    registrant: Registrant
    fund: Fund
    securities: list[BaseInstrument | DebtSecurity | Derivative]

    def __repr__(self):
        return f"{self.registrant.name} [{self.registrant.report_date:%Y-%m-%d}]"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_file(cls, path: str | Path, **kwargs):
        """
        Create NPORT class object from path to filing

        Args:
            path (str or Path):
                Path to the filing
            **kwargs:
                Additional arguments to hand over to the open function

        Returns:
            NPORT: NPORT class object
        """
        with open(path, "r", **kwargs) as file:
            xml = file.read()

        root = BeautifulSoup(xml, features="xml")
        return cls.from_xml(root)

    @classmethod
    def from_str(cls, xml: str):
        """
        Create NPORT class object from filings text

        Args:
            xml (str):
                Text of the filing

        Returns:
            NPORT: NPORT class object
        """
        root = BeautifulSoup(xml, features="xml")
        return cls.from_xml(root)

    @classmethod
    def from_url(cls, url: str):
        """
        Create NPORT class object from filings url

        Args:
            url (str):
                URL of the filing

        Returns:
            NPORT: NPORT class object
        """
        xml = download_text(url)
        return cls.from_str(xml)

    @classmethod
    def from_xml(cls, root: BeautifulSoup):
        """
        Create NPORT class object from the XML root tag

        Args:
            root (BeautifulSoup): Root of the filing

        Returns:
            NPORT: NPORT class object
        """
        form_tag = root.find("formData")

        registrant_tag = form_tag.find("genInfo")
        registrant = Registrant.from_xml(registrant_tag)

        fund_tag = form_tag.find("fundInfo")
        fund = Fund.from_xml(fund_tag)

        securities_tag = form_tag.find("invstOrSecs")
        securities = []
        for security_tag in securities_tag.find_all("invstOrSec"):
            if security_tag.find("debtSec"):
                securities.append(DebtSecurity.from_xml(security_tag, registrant.report_date))
            elif security_tag.find("derivativeInfo"):
                securities.append(Derivative.from_xml(security_tag, registrant.report_date))
            else:
                securities.append(BaseInstrument.from_xml(security_tag, registrant.report_date))

        return cls(
            registrant=registrant,
            fund=fund,
            securities=securities
        )

    def export_prices(self):
        """
        Compile the identifiers and prices for all instruments listed in the report
        """
        price_list = [security.to_list() for security in self.securities]
        return pd.DataFrame(price_list)



import itertools
from typing import Optional, Union, Iterable
from datetime import datetime
import logging
from rich.logging import RichHandler

import pandas as pd
import httpx

from nport._utils import (
    is_valid_filing_date, is_start_of_quarter, filing_date_to_year_quarters,
    current_year_and_quarter, available_quarters, split_filing_date, listify
)
from nport._connector import download_text


index_url = "https://www.sec.gov/Archives/edgar/full-index/{}/QTR{}/{}.{}"
form_specs: list[tuple] = [
    ("form", (0, 12), str),
    ("company", (12, 74), str),
    ("cik", (79, 88), int),
    ("filing_date", (88, 102), str),
    ("accession_number", (102, 156), str)
]
company_specs: list[tuple] = [
    ("company", (0, 62), str),
    ("form", (62, 79), str),
    ("cik", (79, 88), int),
    ("filing_date", (88, 102), str),
    ("accession_number", (102, 141), str)
]
xbrl_specs: list[tuple] = [
    ("cik", 0, int),
    ("company", 1, str),
    ("form", 2, str),
    ("filing_date", 3, str),
    ("accession_number", 4, str)
]

logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)]
)


def read_index_table(index_text: str, index: str) -> pd.DataFrame:
    """
    Read filing index text content into pandas Dataframe
    """
    lines = index_text.rstrip("\n").split("\n")

    # Find where the data starts
    data_start = 0
    for i, line in enumerate(lines):
        if line.startswith("-----"):
            data_start = i + 1
            break

    data = []
    data_lines = lines[data_start:]

    if index.lower() == "xbrl":
        for line in data_lines:
            cells = line.split("|")
            data.append({spec[0]: spec[2](cells[spec[1]]) for spec in xbrl_specs})
    else:
        index_specs = form_specs if index.lower() == "form" else company_specs
        for line in data_lines:
            data.append({spec[0]: spec[2](line[spec[1][0]:spec[1][1]].strip()) for spec in index_specs})
    data_df = pd.DataFrame(data)
    data_df["filing_date"] = pd.to_datetime(data_df["filing_date"], format="%Y-%m-%d")
    data_df["accession_number"] = data_df["accession_number"].str.rsplit("/", expand=False).str[-1]
    data_df["accession_number"] = data_df["accession_number"].str.slice(0, -4)
    return data_df


def fetch_filing_index(year: int, quarter: int, index: str) -> pd.DataFrame | None:
    """
    Download filing index for a specified year and quarter
    """
    url = index_url.format(year, quarter, index, "gz")
    try:
        index_text = download_text(url)
        index_table = read_index_table(index_text, index)
        return index_table
    except httpx.HTTPStatusError as e:
        if is_start_of_quarter() and e.response.status_code == 403:
            logging.info(f"No filings for {year} Q{quarter} since start of new quarter.")
            return None
        logging.error(str(e))
        raise


def get_filings_for_quarters(years_and_quarters: list[tuple[int, int]], index: str
                             ) -> pd.DataFrame | None:
    """
    Fetch the filings for all year, quarter combinations
    """
    filings = []
    for year, quarter in years_and_quarters:
        filings_quarter = fetch_filing_index(year, quarter, index)
        if filings_quarter is not None:
            filings.append(filings_quarter)
    filings = pd.concat(filings)
    return filings


def get_filings(
        year: Optional[int | list[int]] = None,
        quarter: Optional[int | list[int]] = None,
        form: Optional[Union[str, list[str]]] = None,
        filing_date: Optional[str] = None,
        index="form"
) -> pd.DataFrame:
    """
    Downloads the filing index from Edgar and filters by year, quarter, filing_date or form.

    Args:
        year (int or Iterable, optional):
            The year or list of years of the filings, e.g. 2024.
        quarter (int or Iterable, optional):
            The quarter or list of quarters of the filing, e.g. 4 for the forth quarter.
        form (str or Iterable, optional):
            The form or forms of the filing, e.g. "NPORT-P" or ["NPORT-P", "10-K"].
        filing_date (str, optional):
            The filing date to filter by given as single day (YYYY-MM-DD) or a range
            of days (YYYY-MM-DD:YYYY-MM-DD).
        index (str, optional):
            The index type - "form" or "company" or "xbrl". Defaults to "form".

    Returns:
        pd.DataFrame: Dataframe of the queried filings
    """
    if filing_date:
        if not is_valid_filing_date(filing_date):
            logging.warning(
                "Provide a valid filing date in the format YYYY-MM-DD or YYYY-MM-DD:YYYY-MM-DD"
            )
            return pd.DataFrame([])
        year_and_quarters = filing_date_to_year_quarters(filing_date)
    elif not year:
        year, quarter = current_year_and_quarter()
        year_and_quarters = [(year, quarter)]
    else:
        year = listify(year)
        if quarter is None:
            quarter = [1, 2, 3, 4]
        else:
            quarter = listify(quarter)
        available_yq = available_quarters()
        year_and_quarters = [yq for yq in itertools.product(year, quarter) if yq in available_yq]

    if len(year_and_quarters) == 0:
        logging.warning(f"""
Provide a year between 1994 and {datetime.now().year} and optionally a quarter (1-4) for which
the SEC has filings.

    e.g. filings = get_filings(2023) OR
         filings = get_filings(2023, 1)

(You specified the year {year} and quarter {quarter})
        """)
        return pd.DataFrame([])
    filings = get_filings_for_quarters(year_and_quarters, index=index)

    if isinstance(form, str):
        filings = filings.loc[filings["form"] == form.upper()]
    elif isinstance(form, Iterable):
        filings = filings.loc[filings["form"].isin(form)]

    if filing_date:
        start_date, end_date = split_filing_date(filing_date)
        filings = filings.loc[
            (filings["filing_date"] >= start_date) & (filings["filing_date"] <= end_date)
            ]

    filings["url"] = ("https://www.sec.gov/Archives/edgar/data/" + filings["cik"].astype(str) + "/"
                      + filings["accession_number"] + ".txt")
    return filings
